#include "stdafx.h"

#include <assert.h>
#ifdef _DEBUG
#define ASSERT(x) assert(x)
#else
#define ASSERT(x)
#endif

#include "Naa3D_Render.h"

#include "Naa3D_VBallText.h"

#define VBALL_SIZE 16

#define VBALL_TX 256
#define VBALL_TY 256

#define VBALL_NX (VBALL_TX/VBALL_SIZE)
#define VBALL_NY (VBALL_TY/VBALL_SIZE)
#define VBALL_NB (VBALL_NX*VBALL_NY)

static NaaVector3D VBall_Norm[VBALL_SIZE][VBALL_SIZE];
//static int VBall_Mask[VBALL_SIZE];
static Byte VBall_Mask[VBALL_SIZE][VBALL_SIZE];

//#define VBall_GetMask(x,y)   (VBall_Mask[(y)] & (1<<(x)))
//#define VBall_SetMask(x,y)   (VBall_Mask[(y)]|= (1<<(x)))
//#define VBall_ClearMask(x,y) (VBall_Mask[(y)]&=~(1<<(x)))

static void CalcVBallData()
{
	NaaVector3D *pn=VBall_Norm[0];
	Byte        *pm=VBall_Mask[0];
	int x,y;
	float nx,ny,y2,f;
	ny=-1.0f;
	for (y=0;y<VBALL_SIZE;y++)
	{
		y2=sqr(ny);
		nx=1.0f;
		for (x=0;x<VBALL_SIZE;x++)
		{
			f=y2+sqr(nx);
			if (f<1.0f)
			{
				*pm=255;
				pn->x=nx;
				pn->y=ny;
				pn->z=(float)sqrt(1.0f-f);
			}
			else
			{
				f=1.0f/(float)sqrt(f);
				pn->x=nx*f;
				pn->y=ny*f;
				pn->z=0;
			}
			nx-=2.0f/(VBALL_SIZE-1);
			++pn;
			++pm;
		}
		ny+=2.0f/(VBALL_SIZE-1);
	}
}

static Byte VBall_Light[VBALL_SIZE][VBALL_SIZE];

#define VBALL_LIGHT0 64
#define VBALL_DIFFUSE 200
#define VBALL_SPECULAR 100
#define VBALL_SPEC_POW 5

static void CalcVBallLight(NaaVector3D vlight)
{
	int x,y; float f; int l;
	Byte *pdest=VBall_Light[0];
	NaaVector3D *pn=VBall_Norm[0];
	Byte        *pm=VBall_Mask[0];
	for (y=0;y<VBALL_SIZE;y++)
	{
		for (x=0;x<VBALL_SIZE;x++)
		{
			//if (*pm)
			{
				f=-(vlight*(*pn));
				if (f<=0.0f)
					l=VBALL_LIGHT0;
				else
				{
					l=VBALL_LIGHT0+f2int(f*VBALL_DIFFUSE)+f2int((float)pow(f,VBALL_SPEC_POW)*VBALL_SPECULAR);
				}
				if (l<1) l=1;
				else if (l>255) l=255;
				*pdest=(Byte)l;
			}
			//else *pdest=0;
			++pdest;
			++pn;
			++pm;
		}
	}
}

static int nvball_vlight=0;

static NaaVector3D vball_vlight[VBALL_NB];

void NaaVBallTexture::CalcVBallVLight()
{
	int n=0;
	int u,v,face;
	int u2,v2,f2;
	int tx,ty;
	tx=0; ty=0;
	VBallPos pos;

	// on commence par la partie avant
	face=0;
	for (v=FACE_SIZE/2;v<FACE_SIZE;v++)
	{ v2=(FACE_SIZE-1)-v;
		for (u=FACE_SIZE/2;u<FACE_SIZE;u++)
		{ u2=(FACE_SIZE-1)-u;
			vball_vlight[n].x=(u-FACE_SIZE/2)*(2.0f/FACE_SIZE);
			vball_vlight[n].y=(v-FACE_SIZE/2)*(2.0f/FACE_SIZE);
			vball_vlight[n].z=-1.0f;
			pos.t[0].u=(tx  )*(1.0f/VBALL_NX);
			pos.t[0].v=(ty  )*(1.0f/VBALL_NY);
			pos.t[1].u=(tx+1)*(1.0f/VBALL_NX);
			pos.t[1].v=(ty+1)*(1.0f/VBALL_NY);
			tpos[face][v][u]=pos;
			if (u2!=u)
			{
				tpos[face][v][u2].t[0].u=pos.t[1].u;
				tpos[face][v][u2].t[1].u=pos.t[0].u;
				tpos[face][v][u2].t[0].v=pos.t[0].v;
				tpos[face][v][u2].t[1].v=pos.t[1].v;
				if (v2!=v)
				{
					tpos[face][v2][u].t[0].u=pos.t[0].u;
					tpos[face][v2][u].t[1].u=pos.t[1].u;
					tpos[face][v2][u].t[0].v=pos.t[1].v;
					tpos[face][v2][u].t[1].v=pos.t[0].v;

					tpos[face][v2][u2].t[0].u=pos.t[1].u;
					tpos[face][v2][u2].t[1].u=pos.t[0].u;
					tpos[face][v2][u2].t[0].v=pos.t[1].v;
					tpos[face][v2][u2].t[1].v=pos.t[0].v;
				}
			}
			else if (v2!=v)
			{
				tpos[face][v2][u].t[0].u=pos.t[0].u;
				tpos[face][v2][u].t[1].u=pos.t[1].u;
				tpos[face][v2][u].t[0].v=pos.t[1].v;
				tpos[face][v2][u].t[1].v=pos.t[0].v;
			}
			++n;
			++tx; if (tx==VBALL_NX) { tx=0; ++ty; }
		}
	}

  // puis la face arriere
	face=1;
	for (v=FACE_SIZE/2;v<FACE_SIZE;v++)
	{ v2=(FACE_SIZE-1)-v;
		for (u=FACE_SIZE/2;u<FACE_SIZE;u++)
		{ u2=(FACE_SIZE-1)-u;
			vball_vlight[n].x=(u-FACE_SIZE/2)*(2.0f/FACE_SIZE);
			vball_vlight[n].y=(v-FACE_SIZE/2)*(2.0f/FACE_SIZE);
			vball_vlight[n].z=1.0f;
			pos.t[0].u=(tx  )*(1.0f/VBALL_NX);
			pos.t[0].v=(ty  )*(1.0f/VBALL_NY);
			pos.t[1].u=(tx+1)*(1.0f/VBALL_NX);
			pos.t[1].v=(ty+1)*(1.0f/VBALL_NY);
			tpos[face][v][u]=pos;
			if (u2!=u)
			{
				tpos[face][v][u2].t[0].u=pos.t[1].u;
				tpos[face][v][u2].t[1].u=pos.t[0].u;
				tpos[face][v][u2].t[0].v=pos.t[0].v;
				tpos[face][v][u2].t[1].v=pos.t[1].v;
				if (v2!=v)
				{
					tpos[face][v2][u].t[0].u=pos.t[0].u;
					tpos[face][v2][u].t[1].u=pos.t[1].u;
					tpos[face][v2][u].t[0].v=pos.t[1].v;
					tpos[face][v2][u].t[1].v=pos.t[0].v;

					tpos[face][v2][u2].t[0].u=pos.t[1].u;
					tpos[face][v2][u2].t[1].u=pos.t[0].u;
					tpos[face][v2][u2].t[0].v=pos.t[1].v;
					tpos[face][v2][u2].t[1].v=pos.t[0].v;
				}
			}
			else if (v2!=v)
			{
				tpos[face][v2][u].t[0].u=pos.t[0].u;
				tpos[face][v2][u].t[1].u=pos.t[1].u;
				tpos[face][v2][u].t[0].v=pos.t[1].v;
				tpos[face][v2][u].t[1].v=pos.t[0].v;
			}
			++n;
			++tx; if (tx==VBALL_NX) { tx=0; ++ty; }
		}
	}

  // puis les faces haut et bas
	face=3;
	f2=2;
	for (u=0;u<FACE_SIZE;u++)
	{
		tpos[face][0          ][u]=tpos[0][FACE_SIZE-1][u];
		tpos[face][FACE_SIZE-1][u]=tpos[1][FACE_SIZE-1][u];
		tpos[f2  ][0          ][u]=tpos[0][0][u];
		tpos[f2  ][FACE_SIZE-1][u]=tpos[1][0][u];
	}

	for (v=1;v<FACE_SIZE-1;v++)
	{ //v2=(FACE_SIZE-1)-v;
		for (u=FACE_SIZE/2;u<FACE_SIZE;u++)
		{ u2=(FACE_SIZE-1)-u;
			vball_vlight[n].x=(u-FACE_SIZE/2)*(2.0f/FACE_SIZE);
			vball_vlight[n].z=(v-FACE_SIZE/2)*(2.0f/FACE_SIZE);
			vball_vlight[n].y=1.0f;
			pos.t[0].u=(tx  )*(1.0f/VBALL_NX);
			pos.t[0].v=(ty  )*(1.0f/VBALL_NY);
			pos.t[1].u=(tx+1)*(1.0f/VBALL_NX);
			pos.t[1].v=(ty+1)*(1.0f/VBALL_NY);
			tpos[face][v][u]=pos;
			if (u2!=u)
			{
				tpos[face][v][u2].t[0].u=pos.t[1].u;
				tpos[face][v][u2].t[1].u=pos.t[0].u;
				tpos[face][v][u2].t[0].v=pos.t[0].v;
				tpos[face][v][u2].t[1].v=pos.t[1].v;

				tpos[f2][v][u2].t[0].u=pos.t[1].u;
				tpos[f2][v][u2].t[1].u=pos.t[0].u;
				tpos[f2][v][u2].t[0].v=pos.t[1].v;
				tpos[f2][v][u2].t[1].v=pos.t[0].v;
			}

			tpos[f2][v][u].t[0].u=pos.t[0].u;
			tpos[f2][v][u].t[1].u=pos.t[1].u;
			tpos[f2][v][u].t[0].v=pos.t[1].v;
			tpos[f2][v][u].t[1].v=pos.t[0].v;

			++n;
			++tx; if (tx==VBALL_NX) { tx=0; ++ty; }
		}
	}

  // puis les faces du cote
	face=5;
	f2=4;
	for (u=0;u<FACE_SIZE;u++)
	{
		tpos[face][0          ][u]=tpos[0][u][FACE_SIZE-1];
		tpos[face][FACE_SIZE-1][u]=tpos[1][u][FACE_SIZE-1];
		tpos[f2  ][0          ][u]=tpos[0][u][0];
		tpos[f2  ][FACE_SIZE-1][u]=tpos[1][u][0];
	}
	for (v=1;v<FACE_SIZE-1;v++)
	{ //v2=(FACE_SIZE-1)-v;
		for (u=FACE_SIZE/2;u<FACE_SIZE-1;u++)
		{ u2=(FACE_SIZE-1)-u;
			vball_vlight[n].y=(u-FACE_SIZE/2)*(2.0f/FACE_SIZE);
			vball_vlight[n].z=(v-FACE_SIZE/2)*(2.0f/FACE_SIZE);
			vball_vlight[n].x=1.0f;
			pos.t[0].u=(tx  )*(1.0f/VBALL_NX);
			pos.t[0].v=(ty  )*(1.0f/VBALL_NY);
			pos.t[1].u=(tx+1)*(1.0f/VBALL_NX);
			pos.t[1].v=(ty+1)*(1.0f/VBALL_NY);
			tpos[face][v][u]=pos;
			if (u2!=u)
			{
				tpos[face][v][u2].t[0].u=pos.t[0].u;
				tpos[face][v][u2].t[1].u=pos.t[1].u;
				tpos[face][v][u2].t[0].v=pos.t[1].v;
				tpos[face][v][u2].t[1].v=pos.t[0].v;

				tpos[f2][v][u2].t[0].u=pos.t[1].u;
				tpos[f2][v][u2].t[1].u=pos.t[0].u;
				tpos[f2][v][u2].t[0].v=pos.t[1].v;
				tpos[f2][v][u2].t[1].v=pos.t[0].v;
			}

			tpos[f2][v][u].t[0].u=pos.t[1].u;
			tpos[f2][v][u].t[1].u=pos.t[0].u;
			tpos[f2][v][u].t[0].v=pos.t[0].v;
			tpos[f2][v][u].t[1].v=pos.t[1].v;

			++n;
			++tx; if (tx==VBALL_NX) { tx=0; ++ty; }
		}
		// u=FACE_SIZE-1 -> y=1
		tpos[face][v][FACE_SIZE-1]=tpos[3][v][FACE_SIZE-1];
		tpos[face][v][0          ]=tpos[2][v][FACE_SIZE-1];

		tpos[f2][v][FACE_SIZE-1]=tpos[3][v][0];
		tpos[f2][v][0          ]=tpos[2][v][0];

	}

	nvball_vlight=n;
}

static int format_first(int mask)
{
	if (!mask) return 0;
	int f=1;
	while (!(mask&f))
		f<<=1;
	return f;
}

r_val NaaVBallTexture::CalcBitmap(NaaBitmap *b,int format_mask)
{

	CalcVBallData();

	CalcVBallVLight();

	int format;

	if (format_mask&(NaaBitmap::RGBA|NaaBitmap::ABGR))
		format=format_first(format_mask&(NaaBitmap::RGBA|NaaBitmap::ABGR));
	else if (format_mask&(NaaBitmap::A))
		format=NaaBitmap::A;
	else

		format=format_first(format_mask);
	b->Create(VBALL_TX,VBALL_TY,0,format);

	int nc=b->nc;
	int c0=0;
	int alpha=-1;
	if (format==NaaBitmap::RGBA)
	{
		alpha=3;
		--nc;
	}
	else if (format==NaaBitmap::ABGR)
	{
		//alpha=0; ++c0;
		alpha=3;
		--nc;
	}
	int bx,by,n;
	int ix,iy;
	int x,y;

	//float nx,ny;
	//ny=-1.0f;

	Byte *pdest=b->data;

	int dest_dx=b->szPix;
	int dest_dy=b->szLine - dest_dx*VBALL_SIZE;
	n=0;
	for (by=0,iy=0;by<VBALL_NY;by++,iy+=VBALL_SIZE)
	{
		//nx=-1.0f;
		for (bx=0,ix=0;bx<VBALL_NX;bx++,ix+=VBALL_SIZE,n++)
		{
			//NaaVector3D light(nx,ny,-1.0f);
			if (n<nvball_vlight)
			{
				NaaVector3D light=vball_vlight[n];
				light.Normalize();
				CalcVBallLight(light);

				Byte *pl=VBall_Light[0];
				Byte *pm=VBall_Mask[0];

				pdest=b->data+iy*b->szLine+ix*b->szPix;

				for (y=0;y<VBALL_SIZE;y++)
				{
					for (x=0;x<VBALL_SIZE;x++)
					{
						Byte bl=*pl;
						if (alpha>=0)
							pdest[alpha]=(*pm);
						Byte *p=pdest+c0;
						int n=nc;
						do
						{
							*(p++)=bl;
						}
						while (--n);

						pdest+=dest_dx;
						++pl;
						++pm;
					}
					pdest+=dest_dy;
				}
			}
			//nx+=2.0f/(VBALL_NX-1);
		}
		//ny+=2.0f/(VBALL_NY-1);
	}
	return r_ok;
}

NaaVBallTexture::VBallPos * NaaVBallTexture::GetVBallFromLight(NaaVector3D vl)
{
	// on recherche la coordonne la plus grande:
	int face;
	//float u,v,w;
	float x=abs(vl.x);
	float y=abs(vl.y);
	float z=abs(vl.z);
	if (z>=x)
	{ // z >= x
		if (z>=y)
		{ // z >= x && z >= y -> Z
			face=vl.z<0?0:1;
			//z=z;
			x=vl.x;
			y=vl.y;
		}
		else
		{ // z >= x && y > z -> Y
			face=vl.y<0?2:3;
			z=y;
			x=vl.x;
			y=vl.z;
		}
	}
	else
	{ // x > z
		if (x>=y)
		{ // x > z && x >= y -> X
			face=vl.x<0?4:5;
			z=x;
			x=vl.y;
			y=vl.z;
		}
		else
		{ // x > z && y > x -> Y
			face=vl.y<0?2:3;
			z=y;
			x=vl.x;
			y=vl.z;
		}
	}
	ASSERT(z>=abs(x));
	ASSERT(z>=abs(y));
	float fu=(1.0f+x/z)*(FACE_SIZE/2)+0.5f;
	float fv=(1.0f+y/z)*(FACE_SIZE/2)+0.5f;
	int u=(int)fu; //f2int(fu);
	int v=(int)fv; //f2int(fv);
	ASSERT(u>=0 && u<FACE_SIZE);
	ASSERT(v>=0 && v<FACE_SIZE);
	return &(tpos[face][v][u]);
}
